/*

$Header: /sources/tsp/tsp/src/util/libbb/bb_core_posix.c,v 1.4 2007/03/01 18:43:35 deweerdt Exp $

-----------------------------------------------------------------------

TSP Library - core components for a generic Transport Sampling Protocol.

Copyright (c) 2006 Eric NOULARD, Frederik DEWEERDT

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

-----------------------------------------------------------------------

Project   : TSP
Maintainer : tsp@astrium-space.com
Component : Consumer

-----------------------------------------------------------------------

Purpose   : Blackboard POSIX implementation

Note	  : Most of the code was written by Eric, and was historically
	    in bb_core.c
-----------------------------------------------------------------------
 */


#include <errno.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <mqueue.h>		
#include <rtems/rtems/sem.h>	/* a verifier rtems/rtems/sem.h ?? */
#include <pthread.h>
#include <fcntl.h>
#include "bb_core_posix.h"
#include "bb_core.h"
#include "bb_utils.h"

#define LOG(x, ...)



struct bb_used_t bb_used[MAX_NUM_OF_BB];

int bb_count = 0;

/* posix rtems ok */
static int posix_bb_msgq_get(S_BB_T * bb, int create)
{
	char *name_msg;
	int retcode = BB_OK;
	int oflag = 0 ;
//	mode_t mode;
//	struct mq_attr attr;
/* init des attributs a voir  */

// mqd_t mq_open( const char *name, int oflag, mode_t mode, struct mq_attr *attr );

	if (create)
		oflag = O_CREAT | O_EXCL;
		name_msg = bb_utils_build_msg_name(bb->name);
		bb->priv.posix.msg_id = (mqd_t) mq_open( name_msg, oflag | BB_MSG_ACCESS_RIGHT );
	
	if (bb->priv.posix.msg_id < 0) {
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_create", "Cannot message queue %s ", name_msg);
		retcode = BB_NOK;
	}
	free(name_msg);
	return retcode;
}

/* posix rtems ok */
static int posix_bb_lock(volatile S_BB_T * bb)
{
	int32_t retcode;

	assert(bb);

	if (-1 == pthread_mutex_lock( bb->priv.posix.sem_id )) {
		if (EINVAL == errno) {
			LOG(BB_LOG_WARNING, "BlackBoard::bb_lock",
				  "Is BB semaphore destroyed?");
		} else {
			LOG(BB_LOG_SEVERE, "BlackBoard::bb_lock",
				  "sem_open failed");
		}
		retcode = BB_NOK;
	} else {
		retcode = BB_OK;
	}

	return retcode;
}

/* posix rtems ok */
static int posix_bb_unlock(volatile S_BB_T *bb)
{
	int32_t retcode;


	/* shadow do not honor locking
	 * they have neither semaphore nor msgqueue attached
	 */
	if (BB_STATUS_SHADOW == bb->status) {
		return BB_OK;
	}

	if (-1 == pthread_mutex_unlock(bb->priv.posix.sem_id)) {
		if (EINVAL == errno) {
			bb_logMsg(BB_LOG_WARNING, "BlackBoard::bb_unlock",
				  "Is BB semaphore destroyed?");
		}
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_unlock",
			  "semop failed");
		retcode = BB_NOK;
	} else {
		retcode = BB_OK;
	}

	return retcode;
}

/* 
 *	posix_bb_sem_get : posix rtems ok
 */
static int posix_bb_sem_get(S_BB_T * bb, int create)
{
	const pthread_mutexattr_t *attr;
	int32_t retcode;
	char *name_sem;
	
	if (create)
		retcode = pthread_mutexattr_init((pthread_mutexattr_t *) attr);	/* takes defaults attributs */ 
		retcode = pthread_mutexattr_setprotocol((pthread_mutexattr_t *) attr, PTHREAD_PRIO_NONE);
		retcode = pthread_mutex_init(bb->priv.posix.sem_id, attr);

	name_sem = bb_utils_build_sem_name(bb->name);
/*	bb->priv.posix.sem_id = semget(bb_utils_ntok(name_sem), 1, flags | BB_SEM_ACCESS_RIGHT);
	int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *attr ); */
	
	if ( retcode < 0) {
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_create",
		    "Cannot create semaphore <%s key = 0x%x>", name_sem, bb_utils_ntok(name_sem));
		goto err;
	}

/*	u_sem_ctrl.val = 2;
	if (semctl(bb->priv.posix.sem_id, 0, SETVAL, u_sem_ctrl) < 0) {
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_create",
		    "Cannot initialise semaphore <%s key = 0x%x>",
		    name_sem, bb_utils_ntok(name_sem));
		goto err;
	}*/
	/* call semop at least once,
	 * so that we avoid possible race conditions between the
	 * creator and the clients (cf UNPV2 pages 284--285)
	 */
/*	s_semop.sem_num = 0;
	s_semop.sem_op = -1;
	s_semop.sem_flg = 0;
	semop(bb->priv.posix.sem_id, &s_semop, 1);*/
	
	free(name_sem);
	return BB_OK;

      err:
	free(name_sem);
	return BB_NOK;
}

/*
 * Détache le segment de mémoire partagée
 * posix rtems ok 
 */
static int posix_bb_shmem_detach(S_BB_T ** bb)
{
	int32_t retcode;

	retcode = BB_OK;
	assert(bb);
	assert(*bb);

	free(*bb);
	
/*	if (-1 == shmdt((void *) *bb)) {
		LOG(bb_log_severe, "blackboard::bb_detach",
			  "shm detach failed");
		retcode = BB_NOK;
	}*/
	
	*bb = NULL;
	return retcode;
}

/*
 * posix_bb_shmem_attach : posix Rtems ok
 */
static int posix_bb_shmem_attach(S_BB_T ** bb, const char *name)
{
	int i_bb, fd_shm, retcode;
	boolean success = FALSE;

	retcode = BB_OK;
	fd_shm  = 0;

	for (i_bb = 0; i_bb < bb_count; i_bb++){
		if ( strncmp((char *)bb_used[i_bb].name, name, BB_NAME_MAX_SIZE + 1) == 0){
			*bb = bb_used[i_bb].fd;
			success = TRUE;
			retcode = BB_OK;
		}			
	}

	if (!success){
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_attach", "Cannot find the BB named :%s ", name);
		retcode = BB_NOK;
	}
	
	if ((retcode == BB_OK) && ((void *) -1 == *bb)) {
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_attach", "Cannot attach shm segment (%s)");
		retcode = BB_NOK;
	}
	return retcode;
}

/*
 *	posix_bb_shmem_get : posix rtems ok
 */
static int posix_bb_shmem_get(S_BB_T ** bb,
							  const char *name,
							  int n_data,
							  int data_size,
							  int create)
{
	int i_bb, shm_size;
	boolean success = FALSE;
	
	/*shm_name = bb_utils_build_shm_name(name);
	if (!shm_name)
		goto err;*/
	
	shm_size = bb_size(n_data, data_size);
	
	if (create){
		//flags = IPC_CREAT | IPC_EXCL;
		if (bb_count == MAX_NUM_OF_BB-1) {
			LOG(BB_LOG_SEVERE, "BlackBoard::bb_create",
			  "Max number of BB reached !");
			return BB_NOK;			
		}
		
		*bb = (S_BB_T *)malloc(sizeof(S_BB_T));
		
		if (*bb < 0) {
			LOG(BB_LOG_SEVERE, "BlackBoard::bb_create",
			  "Cannot allow memory ");
			*bb = NULL;
			return BB_NOK;
		}
		
		/* created BB memorisation */
		strncpy((char *)bb_used[bb_count].name, name, BB_NAME_MAX_SIZE + 1);
		bb_used[bb_count].fd   = (S_BB_T *)*bb;
		bb_count++;

		/* Init the freshly allocated structure */
		memset(*bb, 0, shm_size);
		(*bb)->bb_version_id = BB_VERSION_ID;
		strncpy((*bb)->name, name,
			BB_NAME_MAX_SIZE + 1);
		(*bb)->max_data_desc_size = n_data;
		(*bb)->data_desc_offset = sizeof(S_BB_T);

		(*bb)->n_data = 0;
		(*bb)->max_data_size = data_size;
		(*bb)->data_offset = (*bb)->data_desc_offset +
		    ((*bb)->max_data_desc_size) * sizeof(S_BB_DATADESC_T);
		(*bb)->data_free_offset = 0;
		(*bb)->status = BB_STATUS_GENUINE;
		(*bb)->priv.posix.shm_size = shm_size;
		(*bb)->priv.posix.fd = *bb;
		(*bb)->type = BB_POSIX;
	}
		
	/* else look for the corresponding fd */
	for (i_bb = 0; i_bb < bb_count; i_bb++){
		if ( strncmp((char *)bb_used[i_bb].name, name, BB_NAME_MAX_SIZE + 1) == 0)
			*bb = bb_used[i_bb].fd;
			success = TRUE;
	}
	
	if (!success){
		LOG(BB_LOG_SEVERE, "BlackBoard::bb_get", "Cannot find the BB named :%s ", name);
		return BB_NOK;	
	}
				  
	return BB_OK;	
}

/*
 *	posix_bb_shmem_destroy
 */
static int32_t posix_bb_shmem_destroy(S_BB_T ** bb)
{
	int32_t retcode = BB_OK;
	int i_bb, j_bb;

	/*S_BB_MSG_T bb_msg;*/

	/* 
	 * On signale la destruction en cours pour les processes qui
	 * resteraient attachés
	 */

	/* FIXME should reserve a message type for signaling destroy */
	/*   bb_msg.mtype = 1; */
	/*   bb_snd_msg(*bb,&bb_msg); */
	/*   sleep(1); */
	
	/* Look for the corresponding BB */
	for (i_bb = 0; i_bb < bb_count; i_bb++){
		if ( (S_BB_T *)bb_used[i_bb].fd == ((S_BB_T *)(*bb)->priv.posix.fd) )
			if ( i_bb == (MAX_NUM_OF_BB - 1)){
				/* cas en fin de tab*/
				bb_used[i_bb].name[0] = '\0'; 
				bb_used[i_bb].fd = NULL;
				bb_count--;
			}
			else { // decallage 
				for (j_bb = i_bb; j_bb < bb_count-1; j_bb++){
					strncpy((char *)bb_used[j_bb].name, (char *)bb_used[j_bb+1].name, BB_NAME_MAX_SIZE + 1);
					bb_used[j_bb].fd = bb_used[j_bb+1].fd;
				}
				bb_count--;
			}			
	}	
	free(*bb);
	*bb = NULL;
	return retcode;
} /* end of posix_bb_shmem_destroy */

/*
 * posix_bb_msgq_recv : posix rtems ok
 */
static int32_t posix_bb_msgq_recv(volatile S_BB_T * bb, S_BB_MSG_T * msg)
{
	int32_t retcode;
	int32_t i_cont;

	retcode = BB_OK;
	assert(bb);

	i_cont = 1;
	/* Réception bloquante */
	while (i_cont) {
		retcode = mq_receive( bb->priv.posix.msg_id, msg->mtext, MAX_SYSMSG_SIZE, NULL );
		
		/* On sort de la boucle si on a pas pris un signal */
		if ((-1 != retcode) || (EINTR != errno)) {
			i_cont = 0;
		}
	}
	if (-1 == retcode) {
		retcode = BB_NOK;
		/* The identifier removed case and INTR
		 * should not generate a log message 
		 * FIXME on devrait pouvoir faire mieux que ça.
		 */
		if ((EIDRM != errno) && (EINTR != errno)) {
			bb_logMsg(BB_LOG_SEVERE,
				  "Blackboard::bb_rcv_msg",
				  "Cannot receive messsage");
		}
	}
	return retcode;
}

/*
 * posix_bb_msgq_send : posix rtems ok
 */
static int32_t posix_bb_msgq_send(volatile S_BB_T * bb, S_BB_MSG_T * msg)
{
	int32_t retcode;
//	struct msqid_ds mystat;
	struct mq_attr mqstat;

	retcode = BB_OK;
	assert(bb);
	/* do not flood message queue if no one read it !! */
//	retcode = msgctl(bb->priv.posix.msg_id, IPC_STAT, &mystat);
	retcode =  mq_getattr( bb->priv.posix.msg_id , &mqstat );

//	if (!(mystat.msg_cbytes > 2 * MAX_SYSMSG_SIZE))
	if ( !(mqstat.mq_curmsgs * mqstat.mq_msgsize > 2 * MAX_SYSMSG_SIZE) ) {
		/* Non blocking send */
		retcode = mq_send( bb->priv.posix.msg_id, msg->mtext, MAX_SYSMSG_SIZE, 0);
		if (-1 == retcode) {
			retcode = BB_NOK;
			if (EAGAIN == errno) {
				/*bb_logMsg(BB_LOG_WARNING,
				   "Blackboard::bb_snd_msg",
				   "Envoi du message impossible (queue pleine) <%s>",
				   syserr);
				 */
			} else {
				LOG(BB_LOG_SEVERE,
				    "Blackboard::bb_snd_msg",
				    "Cannot send message");
			}
		}
	} else {
		/* NOP */
/*     LOG(BB_LOG_WARNING, */
/* 	      "Blackboard::bb_snd_msg", */
/* 	      "Cannot send msg <%d> bytes already queued>",mystat.msg_cbytes); */
	}

	return retcode;
}

/*
 * posix_bb_msgq_isalive : posix rtems ok
 */
static int32_t posix_bb_msgq_isalive(S_BB_T * bb)
{
	int32_t retcode;
	struct mq_attr mqstat;

	retcode =  mq_getattr( bb->priv.posix.msg_id , &mqstat );

	if ( mqstat.mq_curmsgs > 0 ) {
		retcode = BB_NOK;
	} else {
		retcode = BB_OK;
	}

	return retcode;
}

/*
 * posix_bb_sem_destroy : posix rtems ok
 */
static int posix_bb_sem_destroy(S_BB_T *bb)
{
	/* FIXME should the sem be taken before being removed? */
	/* Remove the bb's semaphore */
	//return semctl(bb->priv.posix.sem_id, 0, IPC_RMID);
	return pthread_mutex_destroy( bb->priv.posix.sem_id );
}

/*
 * posix_bb_msgq_destroy : posix rtems ok
 */
static int posix_bb_msgq_destroy(S_BB_T *bb)
{
	char *name_msg;
	name_msg = bb_utils_build_msg_name(bb->name);
	
	/* Destroy the bb's message queue */
	return mq_unlink(name_msg);
}

/*
 * Structure bb_operations : pointeurs de fonctions
 */
struct bb_operations posix_bb_ops = {
	.bb_shmem_get = posix_bb_shmem_get,
	.bb_shmem_attach = posix_bb_shmem_attach,
	.bb_shmem_detach = posix_bb_shmem_detach,
	.bb_shmem_destroy = posix_bb_shmem_destroy,

	.bb_sem_get = posix_bb_sem_get,
	.bb_lock = posix_bb_lock,
	.bb_unlock = posix_bb_unlock,
	.bb_sem_destroy = posix_bb_sem_destroy,

	.bb_msgq_get = posix_bb_msgq_get,
	.bb_msgq_send = posix_bb_msgq_send,
	.bb_msgq_recv = posix_bb_msgq_recv,
	.bb_msgq_isalive = posix_bb_msgq_isalive,
	.bb_msgq_destroy = posix_bb_msgq_destroy,
};
